Additional macros which will probably come in handy in removing code that's been
duplicated between several modules and then updated in only some of them.
--- Linux-PAM-0.75/libpam/include/security/_pam_macros.h	Mon Oct  1 15:57:52 2001
+++ pam-0.75/libpam/include/security/_pam_macros.h	Mon Oct  1 12:06:00 2001
@@ -46,6 +46,24 @@
 	free(reply);                              \
 } while (0)
 
+#ifdef PAM_USER_MATCH
+#ifndef PAM_USER_IN_GROUP
+#define PAM_USER_IN_GROUP
+#endif
+#ifndef PAM_USER_IN_NETGROUP
+#define PAM_USER_IN_NETGROUP
+#endif
+#endif
+
+#ifdef PAM_USER_IN_GROUP
+#ifndef PAM_GETPWNAM_R
+#define PAM_GETPWNAM_R
+#endif
+#ifndef PAM_GETGRNAM_R
+#define PAM_GETGRNAM_R
+#endif
+#endif
+
 /* Wrappers for Unix98's getpwXXX and getgrXXX _r variants, for handling
  * ERANGE error conditions and other weirdness. */
 #define DEFINE_PAM_GETENT(fn,qtype,type) \
@@ -117,6 +135,125 @@
 #ifdef PAM_GETSPNAM_R
 DEFINE_PAM_GETENT(getspnam_r,const char *,struct spwd);
 #endif
+
+#ifdef PAM_USER_IN_GROUP
+static int
+_pam_user_in_group(const char *username, const char *groupname)
+{
+    char *ubuf = NULL, *gbuf = NULL;
+    size_t ubuflen, gbuflen;
+    struct passwd passwd, *pwd;
+    struct group group, *grp;
+    int ret = 0;
+
+    if(_pam_getpwnam_r(username, &passwd, &ubuf, &ubuflen, &pwd) != 0) {
+        pwd = NULL;
+    }
+    if(_pam_getgrnam_r(groupname, &group, &gbuf, &gbuflen, &grp) != 0) {
+        grp = NULL;
+    }
+
+    if((pwd != NULL) && (grp != NULL) && (ret == 0)) {
+        if(pwd->pw_gid == grp->gr_gid) {
+            ret = 2;
+        }
+    }
+#ifdef HAVE_GETGROUPLIST
+    if((pwd != NULL) && (grp != NULL) && (ret == 0)) {
+        gid_t grouplist[64]; /* fixme: use sysconf here */
+        int ngroups, i;
+        ngroups = sizeof(grouplist) / sizeof(grouplist[0]);
+        if(getgrouplist(username, pwd->pw_gid, grouplist, &ngroups) != -1) {
+            for(i = 0; i < ngroups; i++) {
+                if(grouplist[i] == grp->gr_gid) {
+                    ret = 1;
+                    break;
+                }
+            }
+        }
+    }
+#endif
+    if((pwd != NULL) && (grp != NULL) && (ret == 0)) {
+        int i;
+        for(i = 0; (grp->gr_mem != NULL) && (grp->gr_mem[i] != NULL); i++) {
+            if(strcmp(grp->gr_mem[i], pwd->pw_name) == 0) {
+                ret = 1;
+                break;
+            }
+        }
+    }
+    if(ubuf) {
+        free(ubuf);
+    }
+    if(gbuf) {
+        free(gbuf);
+    }
+
+    return ret;
+}
+#endif
+
+#ifdef PAM_USER_IN_NETGROUP
+
+#ifdef HAVE_YP_GET_DEFAULT_DOMAIN
+#include <rpcsvc/ypclnt.h>
+#endif
+
+static int
+_pam_user_in_netgroup(const char *username, const char *netgroupname)
+{
+    int ret = 0;
+#if defined(HAVE_GETNETGRENT) && defined(HAVE_YP_GET_DEFAULT_DOMAIN)
+    char *hostp = NULL, *userp = NULL, *domainp = NULL, *domain = NULL;
+    char hostname[256];
+    int i;
+
+    if((yp_get_default_domain(&domain) != 0) || (domain == NULL)) {
+        return 0;
+    }
+
+    if(setnetgrent(netgroupname) != 0) {
+        return 0;
+    }
+
+    memset(hostname, 0, sizeof(hostname));
+    i = gethostname(hostname, sizeof(hostname) - 1);
+    if((i == 0) || (i == sizeof(hostname) - 1)) {
+        return 0;
+    }
+
+    while(getnetgrent(&hostp, &userp, &domainp) == 0) {
+        if((hostp == NULL) || (strcmp(hostp, hostname) == 0))
+        if((userp == NULL) || (strcmp(userp, username) == 0))
+        if((domainp == NULL) || (strcmp(domainp, domain) == 0)) {
+                ret = 1;
+                break;
+        }
+    }
+    endnetgrent();
+#endif
+    return ret;
+}
+#endif
+
+#ifdef PAM_USER_MATCH
+static int
+_pam_user_match(const char *username, const char *spec)
+{
+    int ret = 0;
+    if(strcmp(username, spec) == 0) {
+        ret = 3;
+    }
+    if((ret == 0) && (spec[0] == '@')) {
+        ret = _pam_user_in_group(username, spec + 1);
+    }
+    if((ret == 0) && (spec[0] == '@')) {
+        ret = _pam_user_in_netgroup(username, spec + 1);
+    }
+    return ret;
+}
+#endif
+
 /* some debugging code */
 
 #ifdef DEBUG
